from typing import Dict, List, Optional, Union

import pydantic
from pydantic import field_validator, model_validator

from truss.base import custom_types, truss_config

DEFAULT_LORA_RANK = 16

# Allowed LoRA rank values for vLLM
ALLOWED_LORA_RANKS = {8, 16, 32, 64, 128, 256, 320, 512}


class SecretReference(custom_types.SafeModelNoExtra):
    name: str


class Compute(custom_types.SafeModelNoExtra):
    node_count: int = 1
    cpu_count: int = 1
    memory: str = "2Gi"
    accelerator: Optional[truss_config.AcceleratorSpec] = None

    def model_dump(self, *args, **kwargs):
        data = super().model_dump(*args, **kwargs)
        if self.accelerator and self.accelerator.accelerator:
            data["accelerator"] = {
                "accelerator": self.accelerator.accelerator.value,
                "count": self.accelerator.count,
            }
        return data

    def to_truss_config(self) -> truss_config.Resources:
        if self.accelerator:
            return truss_config.Resources(
                cpu=str(self.cpu_count),
                memory=self.memory,
                accelerator=self.accelerator,
                node_count=self.node_count,
            )
        return truss_config.Resources(
            cpu=str(self.cpu_count), memory=self.memory, node_count=self.node_count
        )


class CheckpointingConfig(custom_types.SafeModelNoExtra):
    enabled: bool = False
    checkpoint_path: Optional[str] = None


class CacheConfig(custom_types.SafeModelNoExtra):
    enabled: bool = False
    enable_legacy_hf_mount: bool = False


class Runtime(custom_types.SafeModelNoExtra):
    start_commands: List[str] = []
    environment_variables: Dict[str, Union[str, SecretReference]] = {}
    enable_cache: Optional[bool] = None
    checkpointing_config: CheckpointingConfig = CheckpointingConfig()
    cache_config: Optional[CacheConfig] = None

    @model_validator(mode="before")
    @classmethod
    def validate_cache_config(cls, values):
        enable_cache = values.get("enable_cache")
        cache_config = values.get("cache_config")

        if enable_cache is not None and cache_config is not None:
            raise ValueError(
                "Cannot set both 'enable_cache' and 'cache_config'. "
                "'enable_cache' is deprecated. Prefer migrating to 'cache_config' with "
                "`enabled=True` and `enable_legacy_hf_cache=True`."
            )

        # Migrate enable_cache to cache_config if enable_cache is True
        if enable_cache is not None and cache_config is None:
            values["cache_config"] = CacheConfig(
                enabled=enable_cache, enable_legacy_hf_mount=enable_cache
            )

        values.pop(
            "enable_cache", None
        )  # Remove deprecated field or else it will fail server-side validation

        return values


class Image(custom_types.SafeModelNoExtra):
    base_image: str


class TrainingJob(custom_types.SafeModelNoExtra):
    image: Image
    compute: Compute = Compute()
    runtime: Runtime = Runtime()

    def model_dump(self, *args, **kwargs):
        data = super().model_dump(*args, **kwargs)
        data["compute"] = self.compute.model_dump()
        return data


class TrainingProject(custom_types.SafeModelNoExtra):
    name: str
    # TrainingProject is the wrapper around project config and job config. However, we exclude job
    # in serialization so just TrainingProject metadata is included in API requests.
    job: TrainingJob = pydantic.Field(exclude=True)


class Checkpoint(custom_types.SafeModelNoExtra):
    training_job_id: str
    id: str
    name: str
    lora_rank: Optional[int] = (
        None  # lora rank will be fetched through the API if available.
    )

    @field_validator("lora_rank")
    @classmethod
    def validate_lora_rank(cls, v):
        if v is not None and v not in ALLOWED_LORA_RANKS:
            raise ValueError(
                f"lora_rank ({v}) must be one of {sorted(ALLOWED_LORA_RANKS)}"
            )
        return v

    def to_truss_config(self) -> truss_config.Checkpoint:
        return truss_config.Checkpoint(
            id=f"{self.training_job_id}/{self.id}", name=self.id
        )


class CheckpointList(custom_types.SafeModelNoExtra):
    download_folder: str = truss_config.DEFAULT_TRAINING_CHECKPOINT_FOLDER
    base_model_id: Optional[str] = None
    checkpoints: List[Checkpoint] = []

    def to_truss_config(self) -> truss_config.CheckpointList:
        checkpoints = [checkpoint.to_truss_config() for checkpoint in self.checkpoints]
        return truss_config.CheckpointList(
            checkpoints=checkpoints, download_folder=self.download_folder
        )


class DeployCheckpointsRuntime(custom_types.SafeModelNoExtra):
    environment_variables: Dict[str, Union[str, SecretReference]] = {}


class DeployCheckpointsConfig(custom_types.SafeModelNoExtra):
    checkpoint_details: Optional[CheckpointList] = None
    model_name: Optional[str] = None
    deployment_name: Optional[str] = None
    runtime: Optional[DeployCheckpointsRuntime] = None
    compute: Optional[Compute] = None



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, Type, Union


class PatchType(Enum):
    """Types of console requests sent to Django and passed along to pynode."""

    MODEL_CODE = "model_code"
    PYTHON_REQUIREMENT = "python_requirement"
    SYSTEM_PACKAGE = "system_package"
    CONFIG = "config"
    PACKAGE = "package"
    DATA = "data"
    ENVIRONMENT_VARIABLE = "environment_variable"
    EXTERNAL_DATA = "external_data"


class Action(Enum):
    """Types of console requests sent to Django and passed along to pynode."""

    ADD = "ADD"
    UPDATE = "UPDATE"
    REMOVE = "REMOVE"


@dataclass
class PatchBody:
    """Marker class"""

    action: Action

    @abstractmethod
    def to_dict(self):
        pass


@dataclass
class ModelCodePatch(PatchBody):
    path: str  # Relative to model module directory
    content: Optional[str] = None

    def to_dict(self):
        return {"action": self.action.value, "path": self.path, "content": self.content}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return ModelCodePatch(
            action=Action[action_str],
            path=patch_dict["path"],
            content=patch_dict["content"],
        )


@dataclass
class PythonRequirementPatch(PatchBody):
    # For uninstall this should just be the name of the package, but for update
    # should be the full line in the requirements.txt format.
    requirement: str

    def to_dict(self):
        return {"action": self.action.value, "requirement": self.requirement}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return PythonRequirementPatch(
            action=Action[action_str], requirement=patch_dict["requirement"]
        )


@dataclass
class SystemPackagePatch(PatchBody):
    # For uninstall this should just be the name of the package, but for update
    # should be the full line in the requirements.txt format.
    package: str

    def to_dict(self):
        return {"action": self.action.value, "package": self.package}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return SystemPackagePatch(
            action=Action[action_str], package=patch_dict["package"]
        )


@dataclass
class ConfigPatch(PatchBody):
    config: dict
    path: str = "config.yaml"

    def to_dict(self):
        return {"action": self.action.value, "config": self.config, "path": self.path}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return ConfigPatch(
            action=Action[action_str],
            config=patch_dict["config"],
            path=patch_dict["path"],
        )


@dataclass
class DataPatch(PatchBody):
    path: str
    content: Optional[str] = None

    def to_dict(self):
        return {"action": self.action.value, "content": self.content, "path": self.path}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return DataPatch(
            action=Action[action_str],
            content=patch_dict["content"],
            path=patch_dict["path"],
        )


@dataclass
class PackagePatch(PatchBody):
    path: str
    content: Optional[str] = None

    def to_dict(self):
        return {"action": self.action.value, "content": self.content, "path": self.path}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return PackagePatch(
            action=Action[action_str],
            content=patch_dict["content"],
            path=patch_dict["path"],
        )


@dataclass
class EnvVarPatch(PatchBody):
    item: dict

    def to_dict(self):
        return {"action": self.action.value, "item": self.item}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return EnvVarPatch(action=Action[action_str], item=patch_dict["item"])


@dataclass
class ExternalDataPatch(PatchBody):
    item: Dict[str, str]

    def to_dict(self):
        return {"action": self.action.value, "item": self.item}

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return ExternalDataPatch(action=Action[action_str], item=patch_dict["item"])


PATCH_BODY_BY_TYPE: Dict[
    PatchType,
    Type[
        Union[
            ModelCodePatch,
            PythonRequirementPatch,
            SystemPackagePatch,
            ConfigPatch,
            DataPatch,
            PackagePatch,
            EnvVarPatch,
            ExternalDataPatch,
        ]
    ],
] = {
    PatchType.MODEL_CODE: ModelCodePatch,
    PatchType.PYTHON_REQUIREMENT: PythonRequirementPatch,
    PatchType.SYSTEM_PACKAGE: SystemPackagePatch,
    PatchType.CONFIG: ConfigPatch,
    PatchType.DATA: DataPatch,
    PatchType.PACKAGE: PackagePatch,
    PatchType.ENVIRONMENT_VARIABLE: EnvVarPatch,
    PatchType.EXTERNAL_DATA: ExternalDataPatch,
}


@dataclass
class Patch:
    """Request to execute code on console."""

    type: PatchType
    body: PatchBody

    def to_dict(self):
        return {"type": self.type.value, "body": self.body.to_dict()}

    @staticmethod
    def from_dict(patch_dict: Dict):
        typ = PatchType(patch_dict["type"])
        body = PATCH_BODY_BY_TYPE[typ].from_dict(patch_dict["body"])
        return Patch(typ, body)
